@@ -1,13 +1,13 @@
-# $Id: Cache.pm,v 1.3 2002/02/18 18:36:20 darobin Exp $
+# $Id: Cache.pm,v 1.9 2002/12/25 17:59:21 matts Exp $
 
 package Apache::AxKit::Cache;
 use strict;
 
 use Apache;
-use Apache::Constants qw(OK DECLINED);
+use Apache::Constants qw(OK DECLINED SERVER_ERROR);
 use Apache::AxKit::Exception;
 use Digest::MD5 ();
-use Compress::Zlib;
+use Compress::Zlib qw(gzopen);
 use Fcntl qw(:flock O_RDWR O_WRONLY O_CREAT O_RDONLY);
 
 # use vars qw/$COUNT/;
@@ -22,9 +22,10 @@ sub new {
 #        @extras = grep(!/gzip/, @extras);
     }
     
+    local $^W; # suppress "Use of uninitialized value" warnings
     my $key = Digest::MD5->new->add(
             join(':', 
-                $r->get_server_name,
+                $r->hostname,
                 $r->get_server_port,
                 $xmlfile,
                 @extras
@@ -123,14 +124,10 @@ sub write {
     }
     
     if ($self->{gzip} && $AxKit::Cfg->GzipOutput) {
-        AxKit::Debug(3, 'Creating gzip output cache');
-        my $fh = Apache->gensym();
-        if (sysopen($fh, $self->{file}.'new.gz', O_RDWR|O_CREAT)) {
-            flock($fh, LOCK_EX);
-            seek($fh, 0, 0);
-            truncate($fh, 0);
-            print $fh ''.Compress::Zlib::memGzip($_[0]);
-            close $fh;
+        AxKit::Debug(3, "Creating gzip output cache: $self->{file}.gz");
+        if (my $gz = gzopen($self->{file}.'new.gz', "wb")) {
+            $gz->gzwrite($_[0]);
+            $gz->gzclose();
             rename($self->{file}.'new.gz', $self->{file}.'.gz')
                     || throw Apache::AxKit::Exception::IO( -text => "Couldn't rename gzipped cachefile: $!");
         }
@@ -202,7 +199,7 @@ sub get_type {
 
 sub deliver {
     my $self = shift;
-    return if $self->{no_cache};
+    return SERVER_ERROR if $self->{no_cache};
     my $r = $self->{apache};
 
     {
@@ -213,7 +210,7 @@ sub deliver {
             $r->content_type($type);
         }
     }
-    
+
     if ($r->content_type eq 'changeme' && !$r->notes('axkit_passthru_type')) {
         $AxKit::Cfg->AllowOutputCharset(1);
         $r->content_type('text/html; charset=' . ($AxKit::Cfg->OutputCharset || "UTF-8"));
@@ -222,14 +219,13 @@ sub deliver {
         $r->content_type($AxKit::OrigType);
     }
 
-    
+
     my ($transformer, $doit) = AxKit::get_output_transformer();
-    
+
     if ($doit) {
         AxKit::Debug(4, "Cache: Transforming content and printing to browser");
-        $r->send_http_header() unless lc($r->dir_config('Filter')) eq 'on';
-        $r->print( $transformer->( $self->read() ) );
-        return OK;
+        $r->pnotes('xml_string',$self->read());
+        return OK; # upstream deliver_to_browser should handle the rest
     }
     else {
         AxKit::Debug(4, "Cache: Sending untransformed content to browser");
@@ -237,7 +233,7 @@ sub deliver {
         # Make sure we unset PATH_INFO or wierd things can happen!
         $ENV{PATH_INFO} = '';
         $r->path_info('');
-        
+
         if ($self->{gzip} && $AxKit::Cfg->DoGzip) {
             AxKit::Debug(4, 'Cache: Delivering gzipped output');
             $r->filename($self->{file}.'.gz');
@@ -245,10 +241,10 @@ sub deliver {
         else {
             $r->filename($self->{file});
         }
-        
+
         return DECLINED;
     }
-    
+
 }
 
 sub reset {
