@@ -1,4 +1,4 @@
-# $Id: HtmlDoc.pm,v 1.1 2002/03/25 14:06:09 jwalt Exp $
+# $Id: HtmlDoc.pm,v 1.3 2002/06/17 23:15:47 jwalt Exp $
 # Apache::AxKit::Language::HtmlDoc - xhtml->pdf renderer
 package Apache::AxKit::Language::HtmlDoc;
 
@@ -26,7 +26,7 @@ sub handler {
 
     my $parser = XML::LibXML->new();
     local $XML::LibXML::match_cb = \&Apache::AxKit::Language::LibXSLT::match_uri;
-    local $XML::LibXML::open_cb = \&Apache::AxKit::Language::LibXSLT::open_uri;
+    local $XML::LibXML::open_cb = \&Apache::AxKit::Language::LibXSLT::open_content_uri;
     local $XML::LibXML::read_cb = \&Apache::AxKit::Language::LibXSLT::read_uri;
     local $XML::LibXML::close_cb = \&Apache::AxKit::Language::LibXSLT::close_uri;
 
@@ -68,9 +68,9 @@ EOX
     $input =~ s{ href="(?!/|.{0,5}:)}{ href="http://$host$path}g;
     AxKit::Debug(8, "About to shell out to htmldoc - hope you have it installed...");
     AxKit::Debug(10, $input);
-    my $rc = run(['htmldoc','--quiet','--format','pdf13','--truetype','--size','a4','--color','--charset','8859-15','--webpage',$r->dir_config->get('AxHtmlDocOptions'),'-'],\$input,\$result);
+    run(['htmldoc','--quiet','--format','pdf13','--truetype','--size','a4','--color','--charset','8859-15','--webpage',$r->dir_config->get('AxHtmlDocOptions'),'-'],\$input,\$result);
 
-    if (!$rc) {
+    if (substr($result,0,5) ne '%PDF-') {
         throw Apache::AxKit::Exception::Error(-text => 'htmldoc returned error: '.$result);
     }
 
