@@ -1,4 +1,4 @@
-# $Id: LibXSLT.pm,v 1.7 2002/04/22 13:06:39 jwalt Exp $
+# $Id: LibXSLT.pm,v 1.18 2003/01/29 13:51:29 matts Exp $
 
 package Apache::AxKit::Language::LibXSLT;
 
@@ -47,8 +47,9 @@ sub handler {
     }
     
     my $parser = XML::LibXML->new();
+    $parser->expand_entities(1);
     local $XML::LibXML::match_cb = \&match_uri;
-    local $XML::LibXML::open_cb = \&open_uri;
+    local $XML::LibXML::open_cb = \&open_content_uri;
     local $XML::LibXML::read_cb = \&read_uri;
     local $XML::LibXML::close_cb = \&close_uri;
 
@@ -72,11 +73,12 @@ sub handler {
 
     my $stylesheet;
     my $cache = $style_cache{$style->key()};
-    if (!$style->has_changed($cache->{mtime}) && ref($cache->{depends}) eq 'ARRAY') {
+    if (ref($cache) eq 'HASH' && !$style->has_changed($cache->{mtime}) && ref($cache->{depends}) eq 'ARRAY') {
+        AxKit::Debug(8, "[LibXSLT] checking if stylesheet is cached");
         my $changed = 0;
         DEPENDS:
         foreach my $depends (@{ $cache->{depends} }) {
-            my $p = Apache::AxKit::Provider->new($r, key => $depends);
+            my $p = Apache::AxKit::Provider->new_style_provider($r, key => $depends);
             if ( $p->has_changed( $cache->{mtime} ) ) {
                 $changed = 1;
                 last DEPENDS;
@@ -98,14 +100,18 @@ sub handler {
             $style_doc = $parser->parse_fh($fh, $style_uri);
         };
         if ($@) {
-            my $stylestring = ${$style->get_strref()};
-            $style_doc = $parser->parse_string($stylestring, $style_uri);
+            my $stylestring = $style->get_strref();
+            $style_doc = $parser->parse_string($$stylestring, $style_uri);
         }
         
+        local $XML::LibXML::open_cb = \&open_stylesheet_uri;
+        
         $stylesheet = XML::LibXSLT->parse_stylesheet($style_doc);
         
-        $style_cache{$style->key()} = 
+        unless ($r->dir_config('AxDisableXSLTStylesheetCache')) {
+            $style_cache{$style->key()} = 
                 { style => $stylesheet, mtime => time, depends => [ get_depends() ] };
+        }
     }
 
     # get request form/querystring parameters
@@ -115,16 +121,21 @@ sub handler {
 
     my $results = $stylesheet->transform($xml_doc, @params);
     
-    if ($last_in_chain && $XML::LibXSLT::VERSION >= 1.03) {
-        my $encoding = $stylesheet->output_encoding;
-        my $type = $stylesheet->media_type;
-        $r->content_type("$type; charset=$encoding");
+    AxKit::Debug(7, "[LibXSLT] transformation finished, creating $results");
+    
+    if ($last_in_chain) {
+        AxKit::Debug(8, "[LibXSLT] outputting to \$r");
+        if ($XML::LibXSLT::VERSION >= 1.03) {
+            my $encoding = $stylesheet->output_encoding;
+            my $type = $stylesheet->media_type;
+            $r->content_type("$type; charset=$encoding");
+        }
+        $stylesheet->output_fh($results, $r);
     }
 
-    $stylesheet->output_fh($results, $r) if $last_in_chain;
-
+    AxKit::Debug(7, "[LibXSLT] storing results in pnotes(dom_tree) ($r)");
     $r->pnotes('dom_tree', $results);
-
+    
 #         warn "LibXSLT returned $output \n";
 #         print $stylesheet->output_string($results);
     return Apache::Constants::OK;
@@ -144,18 +155,61 @@ sub fixup_params {
 sub match_uri {
     my $uri = shift;
     AxKit::Debug(8, "LibXSLT match_uri: $uri");
+    return 1 if $uri =~ /^(axkit|xmldb):/;
     return $uri !~ /^\w+:/; # only handle URI's without a scheme
 }
 
-sub open_uri {
+sub open_content_uri {
     my $uri = shift || './';
-    AxKit::Debug(8, "LibXSLT open_uri: $uri");
-    my $provider = Apache::AxKit::Provider->new(
-        AxKit::Apache->request(),
-        uri => $uri,
-        );
+    AxKit::Debug(8, "LibXSLT open_content_uri: $uri");
+    
+    if ($uri =~ /^axkit:/) {
+        return AxKit::get_axkit_uri($uri);
+    } elsif ($uri =~ /^xmldb:/) {
+        return Apache::AxKit::Provider::XMLDB::get_xmldb_uri($uri);
+    }
+    
+    # create a subrequest, so we get the right AxKit::Cfg for the URI
+    my $apache = AxKit::Apache->request;
+    my $sub = $apache->lookup_uri(AxKit::FromUTF8($uri));
+    local $AxKit::Cfg = Apache::AxKit::ConfigReader->new($sub);
+    
+    my $provider = Apache::AxKit::Provider->new_content_provider($sub);
+    
     add_depends($provider->key());
     my $str = $provider->get_strref;
+    
+    undef $provider;
+    undef $apache;
+    undef $sub;
+    
+    return $$str;
+}
+
+sub open_stylesheet_uri {
+    my $uri = shift || './';
+    AxKit::Debug(8, "LibXSLT open_stylesheet_uri: $uri");
+    
+    if ($uri =~ /^axkit:/) {
+        return AxKit::get_axkit_uri($uri);
+    } elsif ($uri =~ /^xmldb:/) {
+        return Apache::AxKit::Provider::XMLDB::get_xmldb_uri($uri);
+    }
+    
+    # create a subrequest, so we get the right AxKit::Cfg for the URI
+    my $apache = AxKit::Apache->request;
+    my $sub = $apache->lookup_uri(AxKit::FromUTF8($uri));
+    local $AxKit::Cfg = Apache::AxKit::ConfigReader->new($sub);
+    
+    my $provider = Apache::AxKit::Provider->new_style_provider($sub);
+    
+    add_depends($provider->key());
+    my $str = $provider->get_strref;
+    
+    undef $provider;
+    undef $apache;
+    undef $sub;
+    
     return $$str;
 }
 
